{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# TD 2, Codage de Lempel-Ziv-Welsh\n",
    "*Introduction à la théorie de l’information*\n",
    "\n",
    "##  Algorithme de Lempel-Ziv\n",
    "\n",
    "Grandes lignes de l’algorithme de codage :\n",
    "\n",
    "- on lit le texte jusqu’à trouver un mot m suivi d’une lettre a tels que m soit dans le dictionnaire, mais pas m||a,\n",
    "- on imprime l’indice de m dans le dictionnaire puis la lettre a,\n",
    "- on ajoute m||a dans le dictionnaire,\n",
    "- on continue la lecture après la lettre a \n",
    "\n",
    "Initialement le dictionnaire ne contient que le mot vide.\n",
    "\n",
    "Grandes lignes de l’algorithme de décodage :\n",
    "\n",
    "- on lit un indice i suivi d’une lettre a\n",
    "- on imprime le mot m d’indice i et la lettre a\n",
    "- on ajoute le mot m||a au dictionnaire\n",
    "- on recommence \n",
    "\n",
    "Pour représenter les données :\n",
    "\n",
    "- Pour le codage le dictionnaire sera représenté par un arbre, chaque nœud peut avoir une multitude de fils, un pour chaque lettre. Un mot sera dans le dictionnaire si il existe un nœud (pas forcément une feuille) tel que le chemin de la racine à ce nœud forme le mot en question.\n",
    "- Pour le décodage le dictionnaire sera représenté par une table indexée par des entiers. Chaque nouveau mot sera placé à la fin du tableau. \n",
    "\n",
    "##  Algorithme de Lempel-Ziv-Welsh\n",
    "\n",
    "Grandes lignes de l’algorithme de codage :\n",
    "\n",
    "- on lit le texte jusqu’à trouver un mot m suivi d’une lettre a tels que m soit dans le dictionnaire, mais pas m||a,\n",
    "- on imprime l’indice de m dans le dictionnaire,\n",
    "- on ajoute m||a dans le dictionnaire,\n",
    "- on continue la lecture à partir de la lettre a incluse. \n",
    "\n",
    "Il faudra avoir préalablement placé dans le dictionnaire tous les mots d’une lettre.\n",
    "\n",
    "Grandes lignes de l’algorithme de décodage :\n",
    "\n",
    "- on lit un indice i,\n",
    "- on imprime le mot m d’indice i\n",
    "- on ajoute à la fin du dictionnaire le mot m||a où a est la première lettre du mot suivant (on ne connaîtra a que lorsqu’on aura lu un indice de plus). \n",
    "\n",
    "Comme pour le codage, il faudra avoir préalablement placé dans le dictionnaire tous les mots d’une lettre.\n",
    "\n",
    "Les structures de données pour le codage et le décodage sont identiques à celles de l’algorithme de Lempel-Ziv.\n",
    "\n",
    "##  Questions\n",
    "\n",
    "Un squelette de programme pour les deux algorithmes mentionnés plus haut est donnée dans la cellule ci dessous.\n",
    "\n",
    "Vous pouvez exécuter la cellule et observer les versions compressées pour Lempel-Ziv, celles-ci contiennent des entiers (en base 10) suivis chacun d’une virgule et d’un caractère."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "from math import log, ceil\n",
    "\n",
    "class LZTree:\n",
    "    def __init__(self, i):\n",
    "        self.fils = {}\n",
    "        self.indice = i\n",
    "\n",
    "# lit un entier en base 10, s'arrête à la première virgule\n",
    "# un fichier mal formatté provoquera probablement une erreur\n",
    "def lire_entier_lz(fichier):\n",
    "    c = fichier.read(1)\n",
    "    i = 0\n",
    "    while c != \",\":\n",
    "        if c == '':\n",
    "            return -1\n",
    "        i = 10 * i + int(c)\n",
    "        c = fichier.read(1)\n",
    "    return i\n",
    "\n",
    "def coder_lz(entree,sortie):\n",
    "    f = open(entree, 'r')\n",
    "    g = open(sortie, 'w')\n",
    "    A = racine = LZTree(0)\n",
    "    taille = 1\n",
    "    c = f.read(1)\n",
    "    while c != '':\n",
    "        if c in A.fils:\n",
    "            A = A.fils[c]\n",
    "        else:\n",
    "            g.write(str(A.indice)+\",\" + c)\n",
    "            A.fils[c]=LZTree(taille)\n",
    "            taille +=1\n",
    "            A = racine\n",
    "        c = f.read(1)\n",
    "    g.write(str(A.indice) + \",\")\n",
    "    f.close()\n",
    "    g.close()\n",
    "\n",
    "def decoder_lz(entree,sortie):\n",
    "    f = open(entree, 'r')\n",
    "    g = open(sortie, 'w')\n",
    "    dico = [\"\"]\n",
    "    i = lire_entier_lz(f)\n",
    "    while i != -1:\n",
    "        c = f.read(1)\n",
    "        mot = dico[i] + c\n",
    "        dico.append(mot)\n",
    "        g.write(mot)\n",
    "        i = lire_entier_lz(f)\n",
    "    f.close()\n",
    "    g.close()\n",
    "\n",
    "def coder_lzw(entree,sortie):\n",
    "    f = open(entree, 'r')\n",
    "    g = open(sortie, 'w')\n",
    "    racine = LZTree(-1)\n",
    "    racine.fils = {chr(i):LZTree(i) for i in range(256)}\n",
    "    taille = 256\n",
    "    A = racine.fils[f.read(1)]\n",
    "    c = f.read(1)\n",
    "    while c != '':\n",
    "        if c in A.fils:\n",
    "            A = A.fils[c]\n",
    "        else:\n",
    "            #\n",
    "            # À compléter\n",
    "            #\n",
    "        c = f.read(1)\n",
    "    g.write(str(A.indice) + \",\")\n",
    "    f.close()\n",
    "    g.close()\n",
    "\n",
    "def decoder_lzw(entree,sortie):\n",
    "    f = open(entree, 'r')\n",
    "    g = open(sortie, 'w')\n",
    "    dico = [chr(i) for i in range(256)]\n",
    "    i = lire_entier_lz(f)\n",
    "    g.write(dico[i])\n",
    "    dico.append(dico[i])\n",
    "    i = lire_entier_lz(f)\n",
    "    while i != -1:\n",
    "        dico[-1] += dico[i][0]\n",
    "        g.write(dico[i])\n",
    "        dico.append(dico[i])\n",
    "        i = lire_entier_lz(f)\n",
    "    f.close()\n",
    "    g.close()\n",
    "\n",
    "coder_lz('extrait.txt','extrait.lz')\n",
    "decoder_lz('extrait.lz','extrait.dlz')\n",
    "\n",
    "coder_lz('candide.txt','candide.lz')\n",
    "decoder_lz('candide.lz','candide.dlz')\n",
    "\n",
    "#coder_lzw('extrait.txt','extrait.lzw')\n",
    "#decoder_lzw('extrait.lzw','extrait.dlzw')\n",
    "\n",
    "#coder_lzw('candide.txt','candide.lzw')\n",
    "#decoder_lzw('candide.lzw','candide.dlzw')\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### 3.1   \n",
    "\n",
    "En s’inspirant du codeur de Lempel-Ziv, écrire le codeur Lempel-Ziv-Welsh (i.e. compléter la fonction coder_lzw). Pour ce dernier, le fichier codé contiendra une suite d’entiers (en base 10) séparés par des virgules. On pourra tester la correction grâce au decodeur decoder_lzw.\n",
    "\n",
    "### 3.2   \n",
    "\n",
    "Modifier les programmes de façon à imprimer la longueur que devrait avoir une version binaire du codage. On pourra en particulier comparer les taux de compression.\n",
    "\n",
    "### 3.3   \n",
    "\n",
    "Modifier les programmes de façon à ce que le fichier codé soit binaire (c’est-à-dire composé des caractères ’0’ et ’1’)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
